//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "MakeTriggers.h"

ClassImp(MakeTriggers)

////////////////////////////////////////////////////////////////////////////////////
MakeTriggers::MakeTriggers(const string aStreamName, const int aVerbose): 
Triggers(aVerbose),
Segments(),
Streams(aStreamName,aVerbose){ 
////////////////////////////////////////////////////////////////////////////////////
   
  // default values
  Musername.clear();      // names of user-defined metadata
  Musertype.clear();      // types of user-defined metadata

  // metadata tree
  Mtree = new TTree("metadata", "metadata");
  Mtree->Branch("start",        &Mstart,      "start/D");
  Mtree->Branch("end",          &Mend,        "end/D");
  Mtree->Branch("Mprocessname", &Mprocessname);
  Mtree->Branch("Mstreamname",  &Mstreamname);
  Mtree->Branch("Mprocessversion", &Mprocessversion);
  Mtree->Branch("Mprocessuser", &Mprocessuser);
  Mtree->Branch("Mfmin",        &Mfmin,       "Mfmin/D");
  Mtree->Branch("Mfmax",        &Mfmax,       "Mfmax/D");
  Mtree->Branch("Mqmin",        &Mqmin,       "Mqmin/D");
  Mtree->Branch("Mqmax",        &Mqmax,       "Mqmax/D");
  Mtree->Branch("Msnrmin",      &Msnrmin,     "Msnrmin/D");
  Mtree->Branch("Msnrmax",      &Msnrmax,     "Msnrmax/D");

  // default metadata
  Mprocessname="PROC";
  Mstreamname=Streams::GetName();// FIXME: problem if changed with Streams functions
  Mprocessversion="0";
  Mprocessuser=(string)getenv("USER");
  Mdvar = new double [0];// user-defined
  Mivar = new int    [0];// user-defined
  Msvar = new string [0];// user-defined

  // create the hdf5 memory datatype
  h5_type = new H5::CompType(sizeof(h5_triggers));
  h5_type->insertMember("time", HOFFSET(h5_triggers, h5_time), H5::PredType::NATIVE_DOUBLE);
  h5_type->insertMember("frequency", HOFFSET(h5_triggers, h5_freq), H5::PredType::NATIVE_DOUBLE);
  h5_type->insertMember("tstart", HOFFSET(h5_triggers, h5_tstart), H5::PredType::NATIVE_DOUBLE);
  h5_type->insertMember("tend", HOFFSET(h5_triggers, h5_tend), H5::PredType::NATIVE_DOUBLE);
  h5_type->insertMember("fstart", HOFFSET(h5_triggers, h5_fstart), H5::PredType::NATIVE_DOUBLE);
  h5_type->insertMember("fend", HOFFSET(h5_triggers, h5_fend), H5::PredType::NATIVE_DOUBLE);
  h5_type->insertMember("snr", HOFFSET(h5_triggers, h5_snr), H5::PredType::NATIVE_DOUBLE);
  h5_type->insertMember("q", HOFFSET(h5_triggers, h5_q), H5::PredType::NATIVE_DOUBLE);
  h5_type->insertMember("amplitude", HOFFSET(h5_triggers, h5_amp), H5::PredType::NATIVE_DOUBLE);
  h5_type->insertMember("phase", HOFFSET(h5_triggers, h5_ph), H5::PredType::NATIVE_DOUBLE);

  h5_type_s = new H5::CompType(sizeof(h5_segments));
  h5_type_s->insertMember("start", HOFFSET(h5_segments, h5_start), H5::PredType::NATIVE_DOUBLE);
  h5_type_s->insertMember("end", HOFFSET(h5_segments, h5_end), H5::PredType::NATIVE_DOUBLE);

}

////////////////////////////////////////////////////////////////////////////////////
MakeTriggers::~MakeTriggers(void){
////////////////////////////////////////////////////////////////////////////////////
  if(Verbose>1) cout<<"MakeTriggers::~MakeTriggers"<<endl;
  delete Mtree;
  Musername.clear();
  Musertype.clear();
  delete Mdvar;
  delete Mivar;
  delete [] Msvar;
  delete h5_type;
  delete h5_type_s;
}

////////////////////////////////////////////////////////////////////////////////////
void MakeTriggers::Reset(void){
////////////////////////////////////////////////////////////////////////////////////
  if(Verbose) cout<<"MakeTriggers::Reset: reset all"<<endl;
  Triggers::ResetTriggers();
  Segments::Reset();
  Mtree->Reset();
}

////////////////////////////////////////////////////////////////////////////////////
string MakeTriggers::Write(const string aOutDir,
			   const string aFileFormat,
			   const string aFileName,
			   const string aSubDir,
			   const string aWriteMode,
			   const bool aWriteMeta){
////////////////////////////////////////////////////////////////////////////////////

  // case where there is nothing to do: no trigger, no segment
  if(!Ttree->GetEntries()&&GetNsegments()<1){
    cerr<<"MakeTriggers::Write: there is nothing to write: no trigger, no segment"<<endl;
    return "none";
  }

  // check output directory
  if(!IsDirectory(aOutDir)){
    cerr<<"MakeTriggers::Write: output directory does not exist: "<<aOutDir<<endl;
    return "none";
  }

  // file formats
  vector <string> fileformat;
  if(aFileFormat.find("xml")!=string::npos)  fileformat.push_back("xml");
  if(aFileFormat.find("hdf5")!=string::npos) fileformat.push_back("h5");
  if(aFileFormat.find("txt")!=string::npos)  fileformat.push_back("txt");
  if(aFileFormat.find("root")!=string::npos) fileformat.push_back("root");
  if(!fileformat.size())                     fileformat.push_back("root");

  // no segment -> create one from triggers
  double st, en;
  if(GetNsegments()<1){
    // get time of first and last trigger to create default segment
    st = Ttree->GetMinimum("tstart");
    en = Ttree->GetMaximum("time");
    Segments::AddSegment(floor(st),ceil(en));
  }

  // get segment tree
  TTree *Stree=Segments::GetTree();

  // start - end
  st=Segments::GetStart(0);
  en=Segments::GetEnd(Segments::GetNsegments()-1);
  
  // set internal metadata
  if(Ttree->GetEntries()){
    Mfmin=Ttree->GetMinimum("fstart");
    Mfmax=Ttree->GetMaximum("fend");
    Mqmin=Ttree->GetMinimum("q");
    Mqmax=Ttree->GetMaximum("q");
    Msnrmin=Ttree->GetMinimum("snr");
    Msnrmax=Ttree->GetMaximum("snr");
  }
  else{
    Mfmin   = -1.0;
    Mfmax   = -1.0;
    Mqmin   = -1.0;
    Mqmax   = -1.0;
    Msnrmin = -1.0;
    Msnrmax = -1.0;
  }
  
  // fill metadata tree
  for(int s=0; s<Segments::GetNsegments(); s++){
    Mstart=Segments::GetStart(s);
    Mend=Segments::GetEnd(s);
    Mtree->Fill();
  }

  if(Verbose>1) cout<<"MakeTriggers::Write: "<<Ttree->GetEntries()<<" triggers, "<<Stree->GetEntries()<<" segments"<<endl;

  // output file name: GWOLLUM convention
  ostringstream outfilestream;
  string outfile;
  // loop over output formats
  for(int f=0; f<(int)fileformat.size(); f++){

    // output file name
    if(aFileName.compare("default")) outfilestream<<aOutDir<<"/"<<aFileName<<"."<<fileformat[f];
    else outfilestream<<aOutDir<<"/"<<Streams::GetNameConv()<<"_"<<Mprocessname<<"-"<<(int)st<<"-"<<(int)(ceil(en)-floor(st))<<"."<<fileformat[f];
    outfile = outfilestream.str();
    if(Verbose) cout<<"MakeTriggers::Write: writing "<<outfile<<endl;

    // save trees
    if(!fileformat[f].compare("root")){// ROOT format
      string wm="RECREATE";
      if(aSubDir.compare("")) wm=aWriteMode;// updates are only possible with sub-directories
      TFile outTFile(outfile.c_str(),wm.c_str());
      outTFile.cd();
      if(aSubDir.compare("")){
	outTFile.mkdir(aSubDir.c_str());
	outTFile.cd(aSubDir.c_str());
      }
      Ttree->Write("triggers");
      Stree->Write();
      if(aWriteMeta) Mtree->Write();
      outTFile.Close();
    }
    else if(!fileformat[f].compare("h5"))// HDF5 format
      WriteHDF5(outfile);
    else if(!fileformat[f].compare("xml"))// XML format
      WriteXML(outfile);
    else// TXT format
      WriteTXT(outfile);
        
    outfilestream.clear();
    outfilestream.str("");
  }
  
  delete Stree;
  fileformat.clear();
  Reset();
  
  return outfile;
}

////////////////////////////////////////////////////////////////////////////////////
bool MakeTriggers::InitUserMetaData(vector<string> aMusername, vector<string> aMusertype){
////////////////////////////////////////////////////////////////////////////////////
  if(!aMusername.size()||aMusername.size()!=aMusertype.size()){
    cerr<<"MakeTriggers::InitUserMetaData: the input vectors are not size-compatible"<<endl;
    return false;
  }
 
  if(Verbose) cout<<"MakeTriggers::InitUserMetaData: init user-defined metadata fields"<<endl;

  // get user mdata
  Musername.clear();
  Musertype.clear();
  Musername=aMusername;
  Musertype=aMusertype;

  // allocate memory
  delete Mdvar;
  delete Mivar;
  delete [] Msvar;
  Mdvar = new double  [(int)Musername.size()];
  Mivar = new int     [(int)Musername.size()];
  Msvar = new string  [(int)Musername.size()];

  if(Verbose>1) cout<<"MakeTriggers::InitUserMetaData: meta-data: size = "<<Musername.size()<<endl;
  for(int m=0; m<(int)Musername.size(); m++){
    Mdvar[m]=-1.0;
    Mivar[m]=-1;
    Msvar[m]="none";
    if(Verbose>2)
      cout<<"MakeTriggers::InitUserMetaData: meta-data: name = "<<Musername[m]<<", type = "<<Musertype[m]<<endl;
     
    if(!Musertype[m].compare("d"))      Mtree->Branch(Musername[m].c_str(), &Mdvar[m], (Musername[m]+"/D").c_str());
    else if(!Musertype[m].compare("i")) Mtree->Branch(Musername[m].c_str(), &Mivar[m], (Musername[m]+"/I").c_str());
    else if(!Musertype[m].compare("s")) Mtree->Branch(Musername[m].c_str(), &Msvar[m]);
    else{
      cerr<<"MakeTriggers::InitUserMetaData: the data type "+Musertype[m]+" is not supported"<<endl;
      return false;
    }
  }
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool MakeTriggers::SetUserMetaData(const string aMusername, const double aMuservalue){
////////////////////////////////////////////////////////////////////////////////////
  if(!Musername.size()){
    cerr<<"MakeTriggers::SetUserMetaData: the user meta-data were not defined"<<endl;
    return false;
  }

  for(int m=0; m<(int)Musername.size(); m++){
    if(Musername[m].compare(aMusername)) continue;
    if(Musertype[m].compare("d")){
      cerr<<"MakeTriggers::SetUserMetaData: the data type for "+Musername[m]+" does not match"<<endl;
      return false;
    }
    Mdvar[m]=aMuservalue;
    if(Verbose>1) cout<<"MakeTriggers::SetUserMetaData: set "<<Musername[m]<<"("<<Musertype[m]<<") = "<<Mdvar[m]<<endl;
    break;
  }
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool MakeTriggers::SetUserMetaData(const string aMusername, const int aMuservalue){
////////////////////////////////////////////////////////////////////////////////////
  if(!Musername.size()){
    cerr<<"MakeTriggers::SetUserMetaData: the user meta-data were not defined"<<endl;
    return false;
  }

  for(int m=0; m<(int)Musername.size(); m++){
    if(Musername[m].compare(aMusername)) continue;
    if(Musertype[m].compare("i")){
      cerr<<"MakeTriggers::SetUserMetaData: the data type for "+Musername[m]+" does not match"<<endl;
      return false;
    }
    Mivar[m]=aMuservalue;
    if(Verbose>1) cout<<"MakeTriggers::SetUserMetaData: set "<<Musername[m]<<"("<<Musertype[m]<<") = "<<Mivar[m]<<endl;
    break;
  }
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool MakeTriggers::SetUserMetaData(const string aMusername, const string aMuservalue){
////////////////////////////////////////////////////////////////////////////////////
  if(!Musername.size()){
    cerr<<"MakeTriggers::SetUserMetaData: the user meta-data were not defined"<<endl;
    return false;
  }

  for(int m=0; m<(int)Musername.size(); m++){
    if(Musername[m].compare(aMusername)) continue;
    if(Musertype[m].compare("s")){
      cerr<<"MakeTriggers::SetUserMetaData: the data type for "+Musername[m]+" does not match"<<endl;
      return false;
    }
    Msvar[m]=aMuservalue;
    if(Verbose>1) cout<<"MakeTriggers::SetUserMetaData: set "<<Musername[m]<<"("<<Musertype[m]<<") = "<<Msvar[m]<<endl;
    break;
  }
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool MakeTriggers::WriteHDF5(const string hdf5filename){ 
////////////////////////////////////////////////////////////////////////////////////

  
 
  // create dataset and data space for segments
  h5_segments *h5_s; hsize_t dimsf_s[1];
  dimsf_s[0] = Segments::GetNsegments();
  h5_s = new h5_segments [Segments::GetNsegments()];
  H5::DataSpace dataspace_s(1, dimsf_s);
  for(int s=0; s<Segments::GetNsegments(); s++){
    h5_s[s].h5_start=Segments::GetStart(s);
    h5_s[s].h5_end=Segments::GetEnd(s);
  }

  // create dataset and data space for triggers
  h5_triggers *h5_t; hsize_t dimsf[1];
  if(Ctree->GetEntries()){// use clusters
    dimsf[0] = Ctree->GetEntries();
    h5_t = new h5_triggers [Ctree->GetEntries()];
  }
  else{// use triggers
    dimsf[0] = Ttree->GetEntries();
    h5_t = new h5_triggers [Ttree->GetEntries()];
  }
  H5::DataSpace dataspace(1, dimsf);
 
  // save clusters
  if(Ctree->GetEntries()){
    for(int t=0; t<Ctree->GetEntries(); t++){
      Ctree->GetEntry(t);
      h5_t[t].h5_time=Ctime;
      h5_t[t].h5_freq=Cfreq;
      h5_t[t].h5_snr=Csnr;
      h5_t[t].h5_q=Cq;
      h5_t[t].h5_amp=Camp;
      h5_t[t].h5_ph=Cph;
      h5_t[t].h5_tstart=Ctstart;
      h5_t[t].h5_tend=Ctend;
      h5_t[t].h5_fstart=Cfstart;
      h5_t[t].h5_fend=Cfend;
    }
  }
  // save triggers
  else{
    for(int t=0; t<Ttree->GetEntries(); t++){
      Ttree->GetEntry(t);
      h5_t[t].h5_time=Ttime;
      h5_t[t].h5_freq=Tfreq;
      h5_t[t].h5_snr=Tsnr;
      h5_t[t].h5_q=Tq;
      h5_t[t].h5_amp=Tamp;
      h5_t[t].h5_ph=Tph;
      h5_t[t].h5_tstart=Ttstart;
      h5_t[t].h5_tend=Ttend;
      h5_t[t].h5_fstart=Tfstart;
      h5_t[t].h5_fend=Tfend;
    }
  }
 
  // create the output file
  H5::H5File* h5_file = new H5::H5File(hdf5filename.c_str(), H5F_ACC_TRUNC);

  // Create the data space for the attributes
  H5::DataSpace attr_dataspace = H5::DataSpace(H5S_SCALAR);
  
  // File attributes 
  H5::Attribute attribute;

  H5::StrType strdatatype(H5::PredType::C_S1, 256); // of length 256 characters
  attribute = h5_file->createAttribute("Mprocessname", strdatatype, attr_dataspace);
  attribute.write(strdatatype, Mprocessname);
  attribute = h5_file->createAttribute("Mstreamname", strdatatype, attr_dataspace);
  attribute.write(strdatatype, Mstreamname);
  attribute = h5_file->createAttribute("Mprocessversion", strdatatype, attr_dataspace);
  attribute.write(strdatatype, Mprocessversion);
  attribute = h5_file->createAttribute("Mprocessuser", strdatatype, attr_dataspace);
  attribute.write(strdatatype, Mprocessuser);

  // loop over metadata
  for(int m=0; m<(int)Musername.size(); m++){
    //attribute = h5_file->createAttribute(Musername[m], H5::PredType::STD_I32BE, attr_dataspace);
    if(!Musertype[m].compare("d")){
      attribute = h5_file->createAttribute(Musername[m], H5::PredType::NATIVE_DOUBLE, attr_dataspace);
      attribute.write( H5::PredType::NATIVE_DOUBLE, &(Mdvar[m]));
    }
    else if(!Musertype[m].compare("s")){
      H5::StrType strdatatype(H5::PredType::C_S1, 256); // of length 256 characters
      attribute = h5_file->createAttribute(Musername[m], strdatatype, attr_dataspace);
      attribute.write(strdatatype, Msvar[m]);
    }
    else{
      attribute = h5_file->createAttribute(Musername[m], H5::PredType::NATIVE_INT, attr_dataspace);
      attribute.write( H5::PredType::NATIVE_INT, &(Mivar[m]));
    }
  }

  // create the dataset
  H5::DataSet* dataset;
  dataset = new H5::DataSet(h5_file->createDataSet("triggers", *h5_type, dataspace));

  // Write the data to the dataset
  dataset->write(h5_t, *h5_type);
  delete dataset;

  dataset = new H5::DataSet(h5_file->createDataSet("segments", *h5_type_s, dataspace_s));
  dataset->write(h5_s, *h5_type_s);
  delete dataset;

  // cleaning
  delete h5_file;
  delete h5_t;
  delete h5_s;

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool MakeTriggers::WriteTXT(const string txtfilename){ 
////////////////////////////////////////////////////////////////////////////////////
  
  // open txt file stream        
  ofstream txtfile(txtfilename.c_str());

  // write file header
  txtfile.precision ( 5 );
  txtfile<<"# central time [s]"<<endl;
  txtfile<<"# central frequency [Hz]"<<endl;
  txtfile<<"# snr []"<<endl;
  txtfile<<"# q []"<<endl;
  txtfile<<"# amplitude [Hz^-1/2]"<<endl;
  txtfile<<"# phase [rad]"<<endl;
  txtfile<<"# starting time [s]"<<endl;
  txtfile<<"# ending time [s]"<<endl;
  txtfile<<"# starting frequency [Hz]"<<endl;
  txtfile<<"# ending frequency [Hz]"<<endl;
  
  // print clusters
  if(Ctree->GetEntries()){
    for(int c=0; c<Ctree->GetEntries(); c++){
      Ctree->GetEntry(c);
      txtfile<<fixed<<Ctime<<"\t"<<Cfreq<<"\t"<<Csnr<<"\t"<<Cq<<"\t"<<scientific<<Camp<<"\t"<<fixed<<Cph<<"\t"<<Ctstart<<"\t"<<Ctend<<"\t"<<Cfstart<<"\t"<<Cfend<<endl;
    }
  }

  // print triggers
  else{
    for(int t=0; t<Ttree->GetEntries(); t++){
      Ttree->GetEntry(t);
      txtfile<<fixed<<Ttime<<"\t"<<Tfreq<<"\t"<<Tsnr<<"\t"<<Tq<<"\t"<<scientific<<Tamp<<"\t"<<fixed<<Tph<<"\t"<<Ttstart<<"\t"<<Ttend<<"\t"<<Tfstart<<"\t"<<Tfend<<endl;
    }
  }

  txtfile.close();

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool MakeTriggers::WriteXML(const string xmlfilename){ 
////////////////////////////////////////////////////////////////////////////////////

  // open txt file stream        
  ofstream xmlfile(xmlfilename.c_str());
  xmlfile.precision ( 5 );

  // THIS FUNCTION IS VERY UGLY
  // BUT XML IS AWFUL !!!!
  // THIS OUGHT TO DISAPPEAR
  
  string ifo=Streams::GetDetPrefix();

  //print header
  xmlfile<<"<?xml version='1.0' encoding='utf-8'?>"<<endl;
  xmlfile<<"<LIGO_LW>"<<endl;

  // print process table
  xmlfile<<"\t<Table Name=\"process:table\">"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"process:comment\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"process:node\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"process:domain\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"process:unix_procid\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"process:start_time\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"ilwd:char\" Name=\"process:process_id\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"process:is_online\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"process:ifos\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"process:jobid\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"process:username\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"process:program\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"process:end_time\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"process:version\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"process:cvs_repository\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"process:cvs_entry_time\"/>"<<endl;
  xmlfile<<"\t\t<Stream Delimiter=\",\" Type=\"Local\" Name=\"process:table\">"<<endl;
  xmlfile<<"\t\t\t,,,,,\"process:process_id:0\",,\""<<ifo<<"\",,\""<<Mprocessuser<<"\",\""<<Mprocessname<<"\",,\""<<Mprocessversion<<"\",,"<<endl;
  xmlfile<<"\t\t</Stream>"<<endl;
  xmlfile<<"\t</Table>"<<endl;

  // print process_params table
  xmlfile<<"\t<Table Name=\"process_params:table\">"<<endl;
  xmlfile<<"\t\t<Column Type=\"ilwd:char\" Name=\"process_params:process_id\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"process_params:program\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"process_params:type\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"process_params:value\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"process_params:param\"/>"<<endl;
  xmlfile<<"\t\t<Stream Delimiter=\",\" Type=\"Local\" Name=\"process_params:table\">"<<endl;
  
  for(int m=0; m<(int)Musername.size(); m++){
    if(!Musertype[m].compare("s")) xmlfile<<"\t\t\t\"process:process_id:0\",\""<<Mprocessname<<"\",\"lstring\",\""<<Musername[m]<<"\",\""<<Msvar[m]<<"\"";
    else if(!Musertype[m].compare("i")) xmlfile<<"\t\t\t\"process:process_id:0\",\""<<Mprocessname<<"\",\"int_8s\",\""<<Musername[m]<<"\",\""<<Mivar[m]<<"\"";
    else xmlfile<<"\t\t\t\"process:process_id:0\",\""<<Mprocessname<<"\",\"real_8\",\""<<Musername[m]<<"\",\""<<Mdvar[m]<<"\"";
    if(m!=(int)Musername.size()-1) xmlfile<<","<<endl;
    else xmlfile<<endl;
  }
  xmlfile<<"\t\t</Stream>"<<endl;
  xmlfile<<"\t</Table>"<<endl;

  // print sngl_burst header
  xmlfile<<"\t<Table Name=\"sngl_burst:table\">"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"sngl_burst:ifo\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"sngl_burst:peak_time\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"sngl_burst:peak_time_ns\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"sngl_burst:start_time\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"sngl_burst:start_time_ns\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"real_4\" Name=\"sngl_burst:duration\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"sngl_burst:search\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"ilwd:char\" Name=\"sngl_burst:process_id\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"ilwd:char\" Name=\"sngl_burst:event_id\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"real_4\" Name=\"sngl_burst:peak_frequency\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"real_4\" Name=\"sngl_burst:central_freq\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"real_4\" Name=\"sngl_burst:bandwidth\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"sngl_burst:channel\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"real_4\" Name=\"sngl_burst:amplitude\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"real_4\" Name=\"sngl_burst:snr\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"real_4\" Name=\"sngl_burst:confidence\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"real_8\" Name=\"sngl_burst:chisq\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"real_8\" Name=\"sngl_burst:chisq_dof\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"sngl_burst:param_one_name\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"real_8\" Name=\"sngl_burst:param_one_value\"/>"<<endl;
  xmlfile<<"\t\t<Stream Name=\"sngl_burst:table\" Type=\"Local\" Delimiter=\",\">"<<endl;

  // print clusters
  if(Ctree->GetEntries()){
    // loop over clusters
    for(int c=0; c<Ctree->GetEntries()-1; c++){
      Ctree->GetEntry(c);
      xmlfile<<"\t\t\t\""<<ifo<<"\","<<fixed<<(int)floor(Ctime)<<","<<(int)floor((Ctime-floor(Ctime))*1e9)<<","<<(int)floor(Ctstart)<<","<<(int)floor((Ctstart-floor(Ctstart))*1e9)<<","<<Ctend-Ctstart<<",\""<<Mprocessname<<"\",\"process:process_id:0\",\"sngl_burst:event_id:"<<c<<"\","<<Cfreq<<","<<(Cfend+Cfstart)/2.0<<","<<Cfend-Cfstart<<",\""<<Streams::GetNameSuffix()<<"\","<<scientific<<Camp<<fixed<<","<<Csnr<<",0,0,0,\"phase\","<<Cph<<","<<endl;
    }
    if(Ctree->GetEntries()){
      Ctree->GetEntry(Ctree->GetEntries()-1);
      xmlfile<<"\t\t\t\""<<ifo<<"\","<<fixed<<(int)floor(Ctime)<<","<<(int)floor((Ctime-floor(Ctime))*1e9)<<","<<(int)floor(Ctstart)<<","<<(int)floor((Ctstart-floor(Ctstart))*1e9)<<","<<Ctend-Ctstart<<",\""<<Mprocessname<<"\",\"process:process_id:0\",\"sngl_burst:event_id:"<<Ctree->GetEntries()-1<<"\","<<Cfreq<<","<<(Cfend+Cfstart)/2.0<<","<<Cfend-Cfstart<<",\""<<Streams::GetNameSuffix()<<"\","<<scientific<<Camp<<fixed<<","<<Csnr<<",0,0,0,\"phase\","<<Cph<<endl;
    }
  }

  // print triggers
  else{
    for(int t=0; t<Ttree->GetEntries()-1; t++){
      Ttree->GetEntry(t);
      xmlfile<<"\t\t\t\""<<ifo<<"\","<<fixed<<(int)floor(Ttime)<<","<<(int)floor((Ttime-floor(Ttime))*1e9)<<","<<(int)floor(Ttstart)<<","<<(int)floor((Ttstart-floor(Ttstart))*1e9)<<","<<Ttend-Ttstart<<",\""<<Mprocessname<<"\",\"process:process_id:0\",\"sngl_burst:event_id:"<<t<<"\","<<Tfreq<<","<<(Tfend+Tfstart)/2.0<<","<<Tfend-Tfstart<<",\""<<Streams::GetNameSuffix()<<"\","<<scientific<<Tamp<<fixed<<","<<Tsnr<<",0,0,0,\"phase\","<<Tph<<","<<endl;
    }
    if(Ttree->GetEntries()){
      Ttree->GetEntry(Ttree->GetEntries()-1);
      xmlfile<<"\t\t\t\""<<ifo<<"\","<<fixed<<(int)floor(Ttime)<<","<<(int)floor((Ttime-floor(Ttime))*1e9)<<","<<(int)floor(Ttstart)<<","<<(int)floor((Ttstart-floor(Ttstart))*1e9)<<","<<Ttend-Ttstart<<",\""<<Mprocessname<<"\",\"process:process_id:0\",\"sngl_burst:event_id:"<<Ttree->GetEntries()-1<<"\","<<Tfreq<<","<<(Tfend+Tfstart)/2.0<<","<<Tfend-Tfstart<<",\""<<Streams::GetNameSuffix()<<"\","<<scientific<<Tamp<<fixed<<","<<Tsnr<<",0,0,0,\"phase\","<<Tph<<endl;
    }
  }
  // close sngl_burst table
  xmlfile<<"\t\t</Stream>"<<endl;
  xmlfile<<"\t</Table>"<<endl;
  
  // print segment_definer table
  xmlfile<<"\t<Table Name=\"segment_definer:table\">"<<endl;
  xmlfile<<"\t\t<Column Type=\"ilwd:char\" Name=\"segment_definer:process_id\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"ilwd:char\" Name=\"segment_definer:segment_def_id\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"segment_definer:ifos\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"segment_definer:name\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"segment_definer:comment\"/>"<<endl;
  xmlfile<<"\t\t<Stream Delimiter=\",\" Type=\"Local\" Name=\"segment_definer:table\">"<<endl;
  xmlfile<<"\t\t\t\"process:process_id:0\",\"segment_definer:segment_def_id:0\",\""<<ifo<<"\",\""<<Mprocessname<<"-segments\",\""<<Mprocessname<<" processed segments\""<<endl;
  xmlfile<<"\t\t</Stream>"<<endl;
  xmlfile<<"\t</Table>"<<endl;
  
  // print segment_summary table
  xmlfile<<"\t<Table Name=\"segment_summary:table\">"<<endl;
  xmlfile<<"\t\t<Column Type=\"ilwd:char\" Name=\"segment_summary:process_id\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"segment_summary:start_time\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"segment_summary:start_time_ns\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"segment_summary:end_time\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"segment_summary:end_time_ns\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"ilwd:char\" Name=\"segment_summary:segment_def_id\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"lstring\" Name=\"segment_summary:comment\"/>"<<endl;
  xmlfile<<"\t\t<Stream Delimiter=\",\" Type=\"Local\" Name=\"segment_summary:table\">"<<endl;
  if(Segments::GetNsegments()) xmlfile<<"\t\t\t\"process:process_id:0\","<<fixed<<(int)floor(Segments::GetStart(0))<<","<<(int)floor((Segments::GetStart(0)-floor(Segments::GetStart(0)))*1e9)<<","<<(int)floor(Segments::GetEnd(Segments::GetNsegments()-1))<<","<<(int)floor((Segments::GetEnd(Segments::GetNsegments()-1)-floor(Segments::GetEnd(Segments::GetNsegments()-1)))*1e9)<<",\"segment_summary:segment_def_id:0\",\""<<Mprocessname<<" processed time interval\""<<endl;
  xmlfile<<"\t\t</Stream>"<<endl;
  xmlfile<<"\t</Table>"<<endl;

  // print segments header
  xmlfile<<"\t<Table Name=\"segment:table\">"<<endl;
  xmlfile<<"\t\t<Column Type=\"ilwd:char\" Name=\"segment:process_id\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"ilwd:char\" Name=\"segment:segment_id\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"segment:start_time\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"segment:start_time_ns\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"segment:end_time\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"int_4s\" Name=\"segment:end_time_ns\"/>"<<endl;
  xmlfile<<"\t\t<Column Type=\"ilwd:char\" Name=\"segment:segment_def_id\"/>"<<endl;
  xmlfile<<"\t\t<Stream Name=\"segment:table\" Type=\"Local\" Delimiter=\",\">"<<endl;
     
  for(int s=0; s<Segments::GetNsegments()-1; s++)
    xmlfile<<"\t\t\t\"process:process_id:0\",\"segment:segment_id:"<<s<<"\","<<fixed<<(int)floor(Segments::GetStart(s))<<","<<(int)floor((Segments::GetStart(s)-floor(Segments::GetStart(s)))*1e9)<<","<<(int)floor(Segments::GetEnd(s))<<","<<(int)floor((Segments::GetEnd(s)-floor(Segments::GetEnd(s)))*1e9)<<",\"segment_definer:segment_def_id:0\","<<endl;
  if(Segments::GetNsegments()) xmlfile<<"\t\t\t\"process:process_id:0\",\"segment:segment_id:"<<Segments::GetNsegments()-1<<"\","<<fixed<<(int)floor(Segments::GetStart(Segments::GetNsegments()-1))<<","<<(int)floor((Segments::GetStart(Segments::GetNsegments()-1)-floor(Segments::GetStart(Segments::GetNsegments()-1)))*1e9)<<","<<(int)floor(Segments::GetEnd(Segments::GetNsegments()-1))<<","<<(int)floor((Segments::GetEnd(Segments::GetNsegments()-1)-floor(Segments::GetEnd(Segments::GetNsegments()-1)))*1e9)<<",\"segment_definer:segment_def_id:0\""<<endl;

  // close segments table
  xmlfile<<"\t\t</Stream>"<<endl;
  xmlfile<<"\t</Table>"<<endl;
  
  xmlfile<<"</LIGO_LW>"<<endl;
    
  
  xmlfile.close();

  return true;
}
