//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "EventMap.h"
#include "TApplication.h"

using namespace std;

int main (int argc, char* argv[]){
        
  // check the arguments
  if(argc!=4){
    cerr<<argv[0]<<" usage:"<<endl; 
    cerr<<argv[0]<<" [stop time] [time range] [file root pattern]"<<endl; 
    cerr<<"[stop time]:     stopping GPS time"<<endl; 
    cerr<<"[time range]:    time range of the map [s]"<<endl; 
    cerr<<"[file pattern]:  file pattern (see example)"<<endl; 
    cerr<<endl; 
    cerr<<"Example:"<<endl; 
    cerr<<argv[0]<<" 1025526786 10 /path/to/my/trigger/files/V1:h_16384Hz_"<<endl; 
    cerr<<endl; 
    return 1;
  }

  // get arguments
  double tend = atof(argv[1]);
  double dt = fabs(atof(argv[2]));
  string fileroot = (string)argv[3];
  
  // display
  TApplication *theApp = new TApplication("livemap", &argc, argv);


  // triggers
  EventMap *EM = new EventMap(1, fileroot+"*.root");
  if(EM->GetNTrig()<=0){
    cerr<<argv[0]<<": no triggers"<<endl;
    return 2;
  }

  // timing
  double tstart=EM->GetTimeMax()-dt;
  double t=tstart;
  EM->SetMapTimeRange(0,dt);

  // speed controls
  double tmax;
  int speed=100000;// initial speed
  double time_step=dt/5.0;
  double DT_previous = EM->GetTimeMax()-dt/2-t;
  
  while(t<tend){
    
    // update trigger list
    if(!EM->SetNewPattern(fileroot+"*.root")) break;
    
    // maximum time
    tmax=EM->GetTimeMax()-dt/2;

    // nothing new -> just wait
    if(tmax==t){
      sleep(2);
      continue;
    }

    // adjust speed
    if(tmax-t>1.2*DT_previous) speed=(int)(1.2*(double)speed);// I'm late -> speed up!
    else if(tmax-t<0.8*DT_previous) speed*=(int)(0.8*(double)speed);// I'm going too fast -> slow down
    else speed*=1;// I'm OK -> good speed!
    DT_previous=tmax-t;

    // produce maps
    while(t<tmax){
      EM->BuildMap(0,t);
      EM->PrintMap(0);
      t+=time_step;
      if(t>tend) break;
      if(speed) usleep(speed);
    }

  }
  delete EM;

  return 0;

  theApp->Run(); //<======
  delete theApp;
  return 0;
}

