//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#ifndef __Chain__
#define __Chain__

#include "CUtils.h"
#include "TFile.h"
#include "TChain.h"

using namespace std;

/**
 * Chain ROOT trees in a list of files defined by a file pattern.
 * This class inherits from the <a href="http://root.cern.ch/root/htmldoc/TChain.html">TChain</a> class from ROOT. Only the constructor was modified and a few functions was added.
 * \author    Florent Robinet
 */
class Chain : public TChain {

 public:

  /**
   * @name Constructors and destructors
   @{
  */
  /**
   * Constructor of the Chain class.
   * @param chainname TTree name to be found in the ROOT files.
   * @param pattern file pattern. Any unix-type patterns should be supported (list of files separated by spaces, *, ?)
   * @param verbose activate the verbosity mode if true.
   */
  Chain(const char* chainname, const char* pattern, const bool verbose=false);

  /**
   * Destructor of the Segments class.
   */
  virtual ~Chain(void);
 
  /**
     @}
  */

  /**
   * Adds a file to the chain.
   * The TTree named 'chainname' found in 'afilename' is added at the end of the chain.
   *
   * true is returned if success.
   * @param afilename single file path to be added to the chain.
   */
  bool AddFile(const string afilename);

  /**
   * Returns the number of ROOT files where the designated tree was found.
   */
  inline int GetNfiles(void){ return (int)rootfilename.size(); };

  /**
   * Returns the file path of ROOT file number 'afilenumber'.
   * 
   * "none" is returned if the file cannot be found.
   * @param afilenumber file number in the chain. It must be an integer between 0 and GetNfiles()
   */
  string GetFilePath(const int afilenumber);

  /**
   * Returns the input file pattern provided in the constructor.
   */
  string GetInputFilePattern(void){ return spattern; };

 private:

  vector <string> rootfilename; ///< active root file names
  string spattern;              ///< input file pattern
  string fchainname;            ///< tree name
  bool fverbose;                ///< verbosity flag

  void ChainPOSIX(void);
  void ChainXROOTD(const vector <string> rootfiles);

  ClassDef(Chain,0)
};

#endif


